@@ -35,8 +35,10 @@
 #include "config.h"
 #endif
 
+#include <errno.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <string.h>
 #include <limits.h>
 #include <math.h>
 #include <poll.h>
@@ -73,6 +75,10 @@
 #include <sys/param.h>
 #endif
 
+#if HAVE_SYS_RANDOM_H
+#include <sys/random.h>
+#endif
+
 #if HAVE_GRP_H
 #include <grp.h>
 #endif
@@ -165,6 +171,9 @@ uint32_t dstip;
  */
 static uint8_t dstmac[ETH_ALEN];
 
+static char* payload_suffix = NULL;
+static ssize_t payload_suffix_size = -1;
+
 uint32_t srcip;                   /* autodetected, override with -S/-b/-0 */
 uint8_t srcmac[ETH_ALEN];         /* autodetected, override with -s */
 
@@ -212,6 +221,40 @@ int verbose = 0;  /* Increase with -v */
 /* Doesn't really need to be volatile, but doesn't hurt. */
 static volatile sig_atomic_t time_to_die = 0;
 
+static ssize_t
+xgetrandom(void *buf, const size_t buflen, const unsigned int flags)
+{
+#ifdef HAVE_GETRANDOM
+        return getrandom(buf, buflen, flags);
+#else
+        char* p = buf;
+        for (int n = 0; n < buflen; n++) {
+                p[n] = random() & 0xff;
+        }
+        return buflen;
+#endif
+}
+
+static long int
+xrandom() {
+        const int maxtry = 10;
+        for (int c = 0; c < maxtry; c++) {
+                long int ret;
+                const ssize_t rc = xgetrandom(&ret, sizeof(ret), 0);
+                if (rc == -1) {
+                        fprintf(stderr, "arping: failed to get random bytes: %s\n", strerror(errno));
+                        continue;
+                }
+                if (sizeof(ret) != rc) {
+                        fprintf(stderr, "arping: got too few random bytes %zd, want %zd\n", rc, sizeof(ret));
+                        continue;
+                }
+                return ret;
+        }
+        fprintf(stderr, "arping: failed to get random bytes after %d tries\n", maxtry);
+        exit(1);
+}
+
 /**
  * If possible, chroot.
  *
@@ -245,6 +288,12 @@ drop_fs_root()
                 }
                 return;
         }
+        if (chdir("/")) {
+                if (verbose) {
+                        printf("arping: chdir(/): %s\n", strerror(errno));
+                }
+                return;
+        }
         if (verbose > 1) {
                 printf("arping: Successfully chrooted to %s\n", pw->pw_dir);
         }
@@ -905,7 +954,7 @@ wait_time(double deadline, uint32_t packetwait)
         if (max_wait < 0) {
                 return 0;
         }
-        if (max_wait > packetwait / 1000000) {
+        if (max_wait > packetwait / 1000000.0) {
                 return packetwait;
         }
         return max_wait * 1000000;
@@ -952,8 +1001,6 @@ static char *ts2str(const struct timespec *tv, const struct timespec *tv2,
 	return buf;
 }
 
-
-
 /** Send directed IPv4 ICMP echo request.
  *
  * \param id      IP id
@@ -968,7 +1015,16 @@ pingmac_send(uint16_t id, uint16_t seq)
         // Without this padding some systems (e.g. Raspberry Pi 3
         // wireless interface) failed. dmesg said:
         //   arping: packet size is too short (42 <= 50)
-        const uint8_t padding[16] = {0};
+        const size_t padding_size = sizeof(struct timespec) + payload_suffix_size;
+        uint8_t padding[padding_size];
+        memset(padding, 0, padding_size);
+        {
+                struct timespec ts;
+                getclock(&ts);
+                memcpy(padding, &ts, sizeof(struct timespec));
+                memcpy(&padding[sizeof(struct timespec)],
+                       payload_suffix, payload_suffix_size);
+        }
 
 	int c;
 
@@ -987,7 +1043,8 @@ pingmac_send(uint16_t id, uint16_t seq)
 	}
 
 	if (-1==(ipv4 = libnet_build_ipv4(LIBNET_IPV4_H
-					  + LIBNET_ICMPV4_ECHO_H + 0,
+                                          + LIBNET_ICMPV4_ECHO_H
+                                          + sizeof padding,
 					  0, /* ToS */
 					  id, /* id */
 					  0, /* frag */
@@ -1340,6 +1397,10 @@ pingmac_recv(const char* unused, struct pcap_pkthdr *h, uint8_t *packet)
                 return;
         }
 
+        if (verbose > 3) {
+                printf("arping: ... right dst mac\n");
+        }
+
         // Source MAC must match, if set.
         if (memcmp(dstmac, ethxmas, ETH_ALEN)) {
                 if (memcmp(pkt_srcmac, dstmac, ETH_ALEN)) {
@@ -1347,6 +1408,10 @@ pingmac_recv(const char* unused, struct pcap_pkthdr *h, uint8_t *packet)
                 }
         }
 
+        if (verbose > 3) {
+                printf("arping: ... right src mac\n");
+        }
+
         // IPv4 Address must be me (maybe).
         if (addr_must_be_same) {
                 uint32_t tmp;
@@ -1356,11 +1421,48 @@ pingmac_recv(const char* unused, struct pcap_pkthdr *h, uint8_t *packet)
                 }
         }
 
-        // Must be ICMP echo reply.
+        if (verbose > 3) {
+                printf("arping: ... src IP acceptable\n");
+        }
+
+        // Must be ICMP echo reply type.
         if (htons(hicmp->icmp_type) != ICMP_ECHOREPLY) {
                 return;
         }
 
+        if (verbose > 3) {
+                printf("arping: ... is echo reply type\n");
+        }
+
+        // Must be ICMP echo reply code 0.
+        if (htons(hicmp->icmp_code) != 0) {
+                return;
+        }
+
+        if (verbose > 3) {
+                printf("arping: ... is echo reply code\n");
+        }
+
+        const char* payload = (char*)hicmp + LIBNET_ICMPV4_ECHO_H;
+        const ssize_t payload_size = h->len - (payload - (char*)packet);
+        if (payload_size < 0) {
+                return;
+        }
+        if (payload_size < sizeof(struct timespec) + payload_suffix_size) {
+                return;
+        }
+        if (verbose > 3) {
+                printf("arping: ... correct payload size (%zd)\n",
+                       payload_size);
+        }
+        if (memcmp(&payload[sizeof(struct timespec)],
+                    payload_suffix, payload_suffix_size)) {
+                    return;
+        }
+        if (verbose > 3) {
+                printf("arping: ... correct payload suffix\n");
+        }
+
         update_stats(timespec2dbl(&arrival) - timespec2dbl(&lastpacketsent));
         if (beep) {
                 printf("\a");
@@ -1605,6 +1707,7 @@ arping_main(int argc, char **argv)
         double deadline = -1;
         char bpf_filter[64];
         ebuf[0] = 0;
+        srandom(time(NULL));
 
         for (c = 1; c < argc; c++) {
                 if (!strcmp(argv[c], "--help")) {
@@ -1658,6 +1761,7 @@ arping_main(int argc, char **argv)
 			break;
 		case 'h':
 			usage(0);
+                        exit(0); // Needless but shuts up compiler warnings.
                 case 'g':
                         drop_group = optarg;
                         break;
@@ -1671,7 +1775,9 @@ arping_main(int argc, char **argv)
 					"purpose)\n");
 				exit(1);
 			}
-		case 'I': /* FALL THROUGH */
+			ifname = optarg;
+                        break;
+		case 'I':
 			ifname = optarg;
 			break;
                 case 'm':
@@ -1756,6 +1862,37 @@ arping_main(int argc, char **argv)
                 exit(1);
         }
 
+        // Generate random payload suffix for MAC pings, to be able to
+        // differentiate from unrelated ping replies.
+        if (payload_suffix_size < 0) {
+                payload_suffix_size = 4;
+                payload_suffix = malloc(payload_suffix_size);
+                if (payload_suffix) {
+                        const ssize_t rc = xgetrandom(payload_suffix, payload_suffix_size, 0);
+                        if (rc == -1) {
+                                fprintf(stderr,
+                                        "arping: failed to get %zd random bytes: %s\n",
+                                        payload_suffix_size,
+                                        strerror(errno));
+                                free(payload_suffix);
+                                payload_suffix = NULL;
+                        } else if (payload_suffix_size != rc) {
+                                fprintf(stderr,
+                                        "arping: only got %zd out of %zd bytes for random suffix\n",
+                                        rc, payload_suffix_size);
+                        }
+                } else {
+                        fprintf(stderr, "arping: failed to allocate %zd bytes for payload suffix.\n",
+                                payload_suffix_size);
+                }
+
+                if (!payload_suffix) {
+                        fprintf(stderr, "arping:  Using constant suffix.\n");
+                        payload_suffix = "arping";
+                        payload_suffix_size = strlen(payload_suffix);
+                }
+        }
+
         if (((mode == PINGIP) && opt_T)
             || ((mode == PINGMAC) && (opt_B || dstmac_opt || opt_U))) {
                 fprintf(stderr, "arping: -T can only be used to ping MAC"
@@ -1857,7 +1994,7 @@ arping_main(int argc, char **argv)
 	 * Handle dstip_given instead of ip address after parms (-B really)
 	 */
 	if (mode == NONE) {
-		if (optind + 1 == argc) {
+		if (parm) {
 			mode = is_mac_addr(parm)?PINGMAC:PINGIP;
 		} else if (dstip_given) {
 			mode = PINGIP;
@@ -2098,7 +2235,7 @@ arping_main(int argc, char **argv)
 	} else { /* PINGMAC */
 		int c;
 		for (c = 0; (maxcount < 0 || c < maxcount) && !time_to_die; c++) {
-			pingmac_send(rand(), c);
+			pingmac_send(xrandom(), c);
                         const uint32_t w = wait_time(deadline, packetwait);
                         if (w == 0) {
                                 break;
